# step.py
"""
Step Base Class
---------------
An abstract base class for the steps in the pipeline, including: postprocess,
preprocess, extract and forecaster.
"""
from abc import abstractmethod
import logging

import pandas as pd
from prefect import Task, context
from sklearn.base import BaseEstimator

from soam.cfg import TRACKING_IS_ACTIVE
from soam.utilities.utils import flatten_dict

logger = logging.getLogger(__name__)
try:
    from mlflow import end_run, log_params, start_run
except ModuleNotFoundError:
    logger.debug("Mlflow dependency is not installed.")


class Step(Task, BaseEstimator):
    """
    The base class for all steps.
    All implementations of step have to implement the `run()` method defined below.
    """

    def __init__(self, **kwargs):
        """
        Soam Step init.

        Parameters
        ----------
        kwargs: dict
            extra args.
        """

        super().__init__(**kwargs)

        if TRACKING_IS_ACTIVE:
            self.active_run = None
            self.state_handlers.append(self.set_tracker_run)

    def get_params(self, deep=True):
        out = dict()
        for key in self._get_param_names():
            if key == "savers":
                continue
            value = getattr(self, key)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def get_mlflow_run_name(self):
        return f"{self.__class__.__name__}"

    def set_tracker_run(
        self, obj, old_state, new_state
    ):  # pylint: disable=unused-argument
        """
        Create a new mlflow run.

        Parameters
        ----------
        obj
            the underlying object to which this state handler is attached
        old_state
            the previous state of this object
        new_state
            the proposed new state of this object
        Returns
        -------
        newstate
            the new state of this object.
        """
        if new_state.is_running():
            obj.active_run = start_run(nested=True, run_name=self.get_mlflow_run_name())
            log_params(flatten_dict(self.get_params()))

        if new_state.is_finished():
            end_run()
        return new_state

    @abstractmethod
    def run(self) -> pd.DataFrame:
        """
        Execute this step function

        Parameters
        ----------
        time_series
            A pandas DataFrame containing the data for the step

        Returns
        -------
        pandas.DataFrame
            pandas DataFrame containing the output values.
        """

    def get_task_id(self) -> str:  # pylint: disable=no-self-use
        """
        Obtains the task id.

        Returns
        -------
        flow_run_id
            the id of the flow run.
        """
        return context["flow_run_id"]

    def __repr__(self) -> str:
        return f"<Task: {BaseEstimator.__repr__(self)}>"



"""
Mail Report
-----------
Mail creator and sender. Its a postprocess that sends a report with
the model forecasts.
"""
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import logging
from os.path import basename
from pathlib import Path
import smtplib
from typing import List, Optional, Tuple, Union

from soam.cfg import MAIL_TEMPLATE, get_smtp_cred
from soam.constants import PROJECT_NAME
from soam.core.step import Step

DEFAULT_SUBJECT = "[{end_date}]Forecast report for {metric_name}"
DEFAULT_SIGNATURE = PROJECT_NAME

logger = logging.getLogger(__name__)


class MailReport:
    """
    Builds and sends reports via mail.
    """

    def __init__(
        self,
        mail_recipients_list: List[str],
        metric_name: str,
        setting_path: Optional[str] = None,
    ):
        """
        Create MailReport object.

        Parameters
        ----------
        mail_recipients_list : list of str
            The mails of the recipients for the report.
        metric_name : str
            Name of the metric being forecasted.
        setting_path : str, optional
            The path for the .ini document with the settings.
        """
        self.mail_recipients_list = mail_recipients_list
        credentials = get_smtp_cred(setting_path)
        self.credentials = credentials
        self.metric_name = metric_name

    def send(
        self,
        current_date: str,
        plot_filename: Union[Path, str],
        subject: str = DEFAULT_SUBJECT,
        signature: str = DEFAULT_SIGNATURE,
    ):
        """
        Send email report.

        Parameters
        ----------
        current_date : str
            Date when the report will be sent.
        plot_filename : str or pathlib.Path
             Path of the forecast plot to send.
        subject : str
            Subject of the email.
        signature : str
            Signature for the email.
        """
        logger.info(f"Sending email report to: {self.mail_recipients_list}")

        mime_img, mime_img_name = self._get_mime_images(Path(plot_filename))
        subject, msg_body = self._build_subject_n_msg_body(
            subject, signature, self.metric_name, current_date, mime_img_name
        )

        self._send_mail(
            self.credentials,
            self.mail_recipients_list,
            subject,
            msg_body,
            [mime_img],
            [],
        )

    def _send_mail(
        self,
        smtp_credentials: dict,
        mail_recipients: List[str],
        subject: str,
        mail_body: str,
        mime_image_list: List[MIMEImage],
        attachments: List[str],
    ):
        """
        Send a report email.
        TODO: review method, may be static

        Parameters
        ----------
        smtp_credentials : dict
            Credentials for the SMTP service.
        mail_recipients : list of str
            The mails of the recipients for the report.
            TODO: this data is on self.
        subject : str
            Subject of the email.
        mail_body : str
            The message to be sent
        mime_image_list : list of email.mime.image.MIMEImage
            List of images to sent
        attachments : list of str
            List of files to attach in the email.
        """
        user = smtp_credentials.get("user_address")
        password = smtp_credentials.get("password")
        from_address = smtp_credentials["mail_from"]
        host = smtp_credentials["host"]
        port = smtp_credentials["port"]
        logger.info(
            f"""About to send the following email:
                    'From: ' {from_address}
                    'To: ' {mail_recipients}
                    'Subject: ' {subject}
                    'Using host': {host} and port: {port}"""
        )
        logger.error(f"With the following body: \n {mail_body}")

        msg_root = MIMEMultipart("related")
        msg_root["From"] = from_address
        msg_root["Subject"] = subject
        msg_root["To"] = ", ".join(mail_recipients)

        msg_alt = MIMEMultipart("alternative")
        msg_root.attach(msg_alt)

        msg_text = MIMEText(mail_body, "html")
        msg_alt.attach(msg_text)

        for mim_img in mime_image_list:
            msg_root.attach(mim_img)

        for attachment in attachments:
            with open(attachment, "rb") as f:
                part = MIMEApplication(f.read(), Name=basename(attachment))
            part["Content-Disposition"] = 'attachment; filename="%s"' % basename(
                attachment
            )
            msg_root.attach(part)

        with smtplib.SMTP(host, port) as server:
            server.ehlo()
            if user is not None and password is not None:
                server.starttls()
                server.ehlo()
                server.login(user, password)
            server.sendmail(from_address, mail_recipients, msg_root.as_string())
        logger.info("Email sent succesfully")

    def _build_subject_n_msg_body(
        self, subject: str, signature: str, metric_name: str, end_date, mime_img: str
    ) -> Tuple[str, str]:
        """
        Creates the subject and message body
        TODO: review method, may be static

        Parameters
        ----------
        subject : str
            The subject to format
        signature : str
            The message signature
        metric_name : str
            The name for the metric
        end_date : ?
            ?
        mime_img : str
            The path to the mime image.

        Returns
        -------
        str
            The subject of the email.
        str
            The message body of the email.
        """
        if subject == DEFAULT_SUBJECT:
            subject = subject.format(end_date=end_date, metric_name=metric_name)
        logger.debug(f"Mail subject:\n {subject}")

        jparams = {
            "signature": signature,
            "metric_name": metric_name,
            "end_date": end_date,
            "mime_img": mime_img,
        }
        msg_body = getattr(MAIL_TEMPLATE, "mail_body")(**jparams)
        logger.debug(f"html mail body:\n {msg_body}")
        return subject, msg_body

    def _get_mime_images(self, plot_filename: Path) -> Tuple[MIMEImage, str]:
        """
        Extract images from local dir paths.
        TODO: review method, may be static

        Parameters
        ----------
        plot_filename : pathlib.Path
            The path to the plot image.

        Returns
        -------
        MIMEImage
            The image MIME document.
        str
            The plot filename.
        """
        with plot_filename.open("rb") as img_file:
            msg_image = MIMEImage(img_file.read())
            img_name = str(plot_filename)
            msg_image.add_header("Content-Id", f"<{img_name}>")

        return msg_image, img_name

    def _format_link(self, factor: str) -> str:
        """
        TODO: review unused method

        Parameters
        ----------
        factor

        Returns
        -------

        """
        return f"<a href=#{factor}>{factor}</a>"


class MailReportTask(Step, MailReport):
    """
    Builds the task that sends reports via mail.
    """

    def __init__(self, mail_recipients_list: List[str], metric_name: str, **kwargs):
        """
        Initialization of the Mail Report Task.

        Parameters
        ----------
        mail_recipients_list: List[str]
            List of the recipients of the email to be sent.
        metric_name: str
            Name of the performance metric being measured.
        """
        Step.__init__(self, **kwargs)  # type: ignore
        MailReport.__init__(self, mail_recipients_list, metric_name)

    def run(  # type: ignore
        self,
        current_date: str,
        plot_filename: Union[Path, str],
        subject: str = DEFAULT_SUBJECT,
        signature: str = DEFAULT_SIGNATURE,
    ):
        """
        Run the Mail Report Task.

        Parameters
        ----------
        current_date: str,
            Current datetime as string.
        plot_filename: Union[Path, str],
            The path and filename of the plot.
        subject: str = DEFAULT_SUBJECT,
            The subject for the email.
        signature: str = DEFAULT_SIGNATURE,
            Signature for the email.
        Returns
        -------
        Mail Report Task
            Sends the report via email.
        """
        return self.send(current_date, plot_filename, subject, signature)



"""Anomalies module tests."""

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest

from soam.workflow.anomalies import ConfidenceIntervalAnomaly
from tests.helpers import sample_data_df  # pylint:disable=unused-import

# pylint:disable=redefined-outer-name


@pytest.fixture
def prediction():
    return pd.DataFrame.from_records(
        np.array(
            [
                ('2016-03-01T00:00:00.000000000', 460093, 400000, 500000),
                ('2016-04-01T00:00:00.000000000', 300000, 290000, 400000),
                ('2016-05-01T00:00:00.000000000', 600000, 500000, 700000),
            ],
            dtype=[
                ('ds', '<M8[ns]'),
                ('yhat', '<i8'),
                ('yhat_lower', '<i8'),
                ('yhat_upper', '<i8'),
            ],
        ),
    )


@pytest.fixture
def expected_anomaly_df():
    return pd.DataFrame.from_records(
        np.array(
            [
                (
                    '2016-03-01T00:00:00.000000000',
                    460093,
                    460093,
                    400000,
                    500000,
                    False,
                    False,
                ),
                (
                    '2016-04-01T00:00:00.000000000',
                    450935,
                    300000,
                    290000,
                    400000,
                    False,
                    True,
                ),
                (
                    '2016-05-01T00:00:00.000000000',
                    471421,
                    600000,
                    500000,
                    700000,
                    True,
                    False,
                ),
            ],
            dtype=[
                ('ds', '<M8[ns]'),
                ('y', '<i8'),
                ('yhat', '<i8'),
                ('yhat_lower_y', '<i8'),
                ('yhat_upper_y', '<i8'),
                ('outlier_lower_y', 'bool'),
                ('outlier_upper_y', 'bool'),
            ],
        ),
    )


def test_confidence_interval_anomaly(sample_data_df, prediction, expected_anomaly_df):
    """ Test confidence interval based anomaly detector.
        Test that different outlier cases match with expected outputs.
    """
    detector = ConfidenceIntervalAnomaly(metric="y")
    anomaly_df = detector.run(time_series=sample_data_df, prediction=prediction)
    assert_frame_equal(anomaly_df, expected_anomaly_df)


def test_metric_not_present(sample_data_df, prediction):
    """Test that the ci_anomaly task raises appropiate exception on missing metric."""
    detector = ConfidenceIntervalAnomaly(metric="gmv")
    with pytest.raises(ValueError):
        detector.run(time_series=sample_data_df, prediction=prediction)



"""
Anomaly detection module.
"""
import logging
from typing import NamedTuple  # pylint:disable=unused-import

import pandas as pd

from soam.constants import DS_COL, YHAT_COL
from soam.core import Step

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class IntervalColumns(NamedTuple):
    lower: str
    upper: str


class ConfidenceIntervalAnomaly(Step):
    def __init__(  # type: ignore
        self,
        metric: str,
        ds_col: str = DS_COL,
        response_col: str = YHAT_COL,
        interval_cols: IntervalColumns = None,
        **kwargs,
    ):
        """Detect anomaly of given value and its boundaries.

        Parameters
        ----------
        metric : str
            metric name to compare to bounds
        ds_col : str, optional
            Date column, by default DS_COL
        keep_cols_pred : Union[str, List[str]], optional
            Name of the prediction column, by default []
        interval_cols : List[str], optional
            Column names for prediction boundaries, by default []
        """
        super().__init__(**kwargs)
        self.metric = metric
        self.time_series = pd.DataFrame()
        self.prediction = pd.DataFrame()
        self.interval_cols = interval_cols or IntervalColumns(
            lower=f"{YHAT_COL}_lower", upper=f"{YHAT_COL}_upper"
        )
        self.response_col = response_col
        self.ds_col = ds_col

    def run(  # type: ignore
        self, prediction: pd.DataFrame, time_series: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Detect anomaly with forecasted boundaries.

        Parameters
        ----------
            prediction: pandas.DataFrame
                containing as minimum the first column
                with DataTime values, and the boundaries columns defined by
                interval_cols
            time_series: pandas.DataFrame
                containing as minimum the first column
                with DataTime values, the second column the y to predict
                and the other columns more data
        Returns
        -------
        pandas.DataFrame
            A pandas DataFrame with the outliers columns.
        """
        prediction = prediction[[self.ds_col, self.response_col, *self.interval_cols]]
        prediction = prediction.astype({self.ds_col: "datetime64[ns]"})

        if self.metric not in time_series.columns:
            raise ValueError(f"Metric {self.metric} not present in time_series.")

        time_series = time_series.iloc[-len(prediction) :]
        time_series = time_series.astype({self.ds_col: "datetime64[ns]"})
        outlier = pd.merge_asof(time_series, prediction, on=self.ds_col)
        outlier["default"] = False
        outlier[f"outlier_lower_{self.metric}"] = outlier.default.where(
            outlier[self.interval_cols.lower] < outlier[self.metric], True
        )
        outlier[f"outlier_upper_{self.metric}"] = outlier.default.where(
            outlier[self.interval_cols.upper] > outlier[self.metric], True
        )
        outlier.drop("default", axis=1, inplace=True)

        metric_lower = f"{self.interval_cols.lower}_{self.metric}"
        metric_upper = f"{self.interval_cols.upper}_{self.metric}"
        outlier = outlier.rename(columns={self.interval_cols.lower: metric_lower})
        outlier = outlier.rename(columns={self.interval_cols.upper: metric_upper})
        logger.info(  # pylint: disable=logging-format-interpolation
            f"""Anomalies calculated for metric: {self.metric}
            with values: y={outlier[self.metric].iloc[0]:.3f},
            prediction_lower={outlier[metric_lower].iloc[0]:.3f},
            prediction_upper={outlier[metric_upper].iloc[0]:.3f}"""
        )
        return outlier


__all__ = ['ConfidenceIntervalAnomaly']



"""Prophet model wrapper."""
import logging
from typing import Dict, List, Union

import pandas as pd
from sklearn.base import BaseEstimator

from soam.constants import DS_COL
from soam.models.base import SkWrapper, sk_constructor_wrapper
from soam.utilities.utils import SuppressStdOutStdErr

# pylint:disable=super-init-not-called, no-member

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

try:
    from fbprophet import Prophet
except ImportError:
    logger.warning("No Prophet support")
    logger.warning("If you want to use it, ´pip install soam[prophet]´")


class SkProphet(SkWrapper):
    """Scikit-Learn Prophet model wrapper."""

    @sk_constructor_wrapper(Prophet)
    def __init__(
        self,
        extra_seasonalities: List[Dict] = None,
        extra_regressors: List[Union[str, Dict]] = None,
        fit_params: Dict = None,
        ds_col: str = DS_COL,
        full_output: bool = False,
    ):
        """Constructor with extra parameters in addition to the Prophet ones.

        Parameters
        ----------
        extra_seasonalities : List[Dict], optional
            Extra seasonalities for Prophet model, by default None
        extra_regressors : List[Union[str, Dict]], optional
            Extra regressor columns for Prophet model, by default None
        fit_params : Dict, optional
            Parameters for Prophet's fit() call, by default None
        ds_col : str, optional
            Date column name, by default None
        full_output : bool, default False
            Return full Prophet output or just prediction column.
        """
        self.extra_seasonalities = extra_seasonalities
        self.extra_regressors = extra_regressors
        self.fit_params = fit_params
        self.ds_col = ds_col
        self.full_output = full_output
        self.model = BaseEstimator()

    def fit(self, X: pd.DataFrame, y: pd.Series):
        """Fit estimator to data."""
        self.model = self._init_sk_model(Prophet, clean=True)
        self._add_extra_params()
        df = self._transform_to_input_format(X, y)
        with SuppressStdOutStdErr():
            if self.fit_params is None:
                self.fit_params = {}
            self.model.fit(df, **self.fit_params)
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's predict."""
        X = self._transform_to_input_format(X)
        predictions = self.model.predict(X)
        predictions = self._transform_to_output_format(predictions)
        return predictions

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's transform"""
        return self.predict(X)

    def fit_transform(self, X: pd.DataFrame, y: pd.Series):
        """Scikit learn's fit_transform"""
        self.fit(X, y)
        return self.transform(X)

    def _transform_to_output_format(self, predictions: pd.Series) -> pd.DataFrame:
        """Transform Prophet output to SoaM format."""
        predictions.reset_index(drop=True, inplace=True)
        if self.ds_col != DS_COL:
            predictions.rename({"ds": self.ds_col}, axis=1, inplace=True)
        if self.full_output:
            return predictions
        return predictions[[self.ds_col, "yhat"]]

    def _transform_to_input_format(
        self, X: pd.DataFrame, y: pd.Series = None
    ) -> pd.DataFrame:
        """Transform input to Prophet compatible df."""
        if self.ds_col != DS_COL:
            X.rename({self.ds_col: "ds"}, axis=1, inplace=True)
        if y is not None:
            return X.assign(**{"y": y})
        return X

    def _add_extra_params(self) -> None:
        """Add regressors and seasonalities to Prophet model."""
        if self.extra_regressors is not None:
            for regressor in self.extra_regressors:
                if isinstance(regressor, str):
                    self.model.add_regressor(regressor)
                else:
                    self.model.add_regressor(**regressor)

        if self.extra_seasonalities is not None:
            for seasonality in self.extra_seasonalities:
                self.model.add_seasonality(**seasonality)



from unittest.mock import patch

import mlflow

from soam.core import SoamFlow
from soam.workflow import Slicer
from tests.helpers import sample_data_df  # pylint: disable=unused-import


def test_simple_flow(sample_data_df, tmpdir):  # pylint: disable=redefined-outer-name
    tmp_path = "file://" + str(tmpdir) + "/mlruns"
    with patch("soam.core.runner.TRACKING_URI", tmp_path), patch(
        "soam.core.runner.TRACKING_IS_ACTIVE", True
    ), patch("soam.core.step.TRACKING_IS_ACTIVE", True):
        df = sample_data_df
        df['metric'] = 1
        dimensions = ["y"]
        ds_col = 'ds'
        metrics = ['metric']
        slice_task = Slicer(ds_col=ds_col, dimensions=dimensions, metrics=metrics)
        with SoamFlow(name="flow") as flow:
            _ = slice_task(sample_data_df)
        flow.run()
        log_df = mlflow.search_runs(['0'])
        assert len(log_df) == 2
        assert log_df['tags.mlflow.runName'].tolist() == ['Slicer', 'flow_run']
        slicer_logs = log_df[log_df['tags.mlflow.runName'] == 'Slicer'].iloc[0]
        assert slicer_logs['params.dimensions'] == str(dimensions)
        assert slicer_logs['params.metrics'] == str(metrics)
        assert slicer_logs['params.ds_col'] == str(ds_col)
